(ns puppetlabs.rbac.services.rbac
 (:require [puppetlabs.trapperkeeper.core :refer [defservice]]
           [puppetlabs.trapperkeeper.services :refer [service-context]]
           [puppetlabs.rbac.db-utils :refer [migrate!]]
           [puppetlabs.rbac.shiro-utils :as shiro]
           [puppetlabs.rbac.storage :as st]
           [puppetlabs.rbac.storage.permissioned :as pm]
           [puppetlabs.rbac.sync :as sn]
           [clj-time.core :refer [now plus hours after?]]
           [clj-jwt.core :as jwt]
           [clj-jwt.key :as jwt-key]
           [clj-jwt.intdate :refer [intdate->joda-time]]
           [clojure.tools.logging :as log]))

(defprotocol RbacService
  "Protocol spec for Authentication and Authorization TK service"
  (get-subject                [this column value])
  (update-local-user-password! [this login new-password clear-token-uuid?])
  (password-matches?          [this login old-password])
  (reset-password-token!      [this requesting-login sid])
  (validate-reset-token       [this token])
  (get-directory-settings     [this subject])
  (update-directory-settings! [this subject data])
  (get-types                  [this])
  (session-id->subject        [this session-id])
  (login->subject             [this login])
  (security-manager           [this]))

(defn do-update-directory-settings!
  "This function updates the directory settings in the database
  and then refreshes the provided shiro realm with the new settings.
  The version that requires a subject argument is the one meant for exposure
  in any API; the version that does not should only be used internally."
  ([db ds-realm settings]
   (st/update-directory-settings! db settings)
   (let [ds-settings (st/get-directory-settings db)]
     (shiro/connect-realm ds-settings ds-realm)))
  ([subject db ds-realm settings]
   (pm/update-directory-settings! subject db settings)
   (let [ds-settings (st/get-directory-settings db)]
     (shiro/connect-realm ds-settings ds-realm))))

(def password-service (org.apache.shiro.authc.credential.DefaultPasswordService.))

(defn hash-password [clear-text]
  (.encryptPassword password-service clear-text))

(defn create-token-map
  "Create a Clojure representation of a JWT token claim:
     :iss - issuer - in this case the admin user requesting the token
     :sub - subject - UUID representing the user having their password reset
     :exp - expiration - the expiration date of the token
     :iat - issued at - time the token was used"
  [password-reset-expiration requesting-login reset-uuid]
  (let [current-time (now)]
    {:iss requesting-login
     :sub reset-uuid
     :exp (plus current-time (hours password-reset-expiration))
     :iat current-time}))

(defn create-password-reset-token
  "Create a JWT token string that encodes an expiration time of
   `password-reset-expiration` hours from now. Signs the token
   so it cannot be tampered with"
  [key password-reset-expiration requesting-login reset-uuid]
  (let [current-time (now)]
    (-> (create-token-map password-reset-expiration requesting-login reset-uuid)
        jwt/jwt
        (jwt/sign :RS256 key)
        jwt/to-str)))

(defn expired-token? [token]
  (let [expiry-date (intdate->joda-time (get-in token [:claims :exp]))]
    (after? (now) expiry-date)))

(defn verify-token [token public-cert]
  (try
    (jwt/verify token public-cert)
    (catch Exception e
      (log/info "Error verifying token" e)
      false)))

(def tampered-token? (complement verify-token))

(defn validate-token [rbac-svc token public-cert]
  (cond
   (tampered-token? token public-cert)
   "Password reset failed, check to ensure the full token text was included"

   (expired-token? token)
   "Password reset failed due to expired token"

   :else
   nil))

(defn minutes->millis [x]
  (* 60 1000 x))

(defservice rbac-service
  RbacService
  [[:ConfigService get-in-config]]
  (init [this context]
        (let [db (get-in-config [:rbac-database])
              rbac-config (get-in-config [:webserver :rbac])
              global-config (get-in-config [:global])]
          (migrate! db)
          (let [ds-cfg     (st/get-directory-settings db)
                db-conn-pool (shiro/get-data-source db)
                jdbc-realm (shiro/create-jdbc-realm db-conn-pool)
                ds-realm   (shiro/create-ds-realm ds-cfg)
                sec-mgr    (shiro/create-security-manager [jdbc-realm ds-realm]
                                                          (minutes->millis (:session-timeout global-config 60)))
                host-private-key (jwt-key/private-key (:ssl-key rbac-config))
                host-public-key (jwt-key/public-key (:ssl-cert rbac-config))
                password-reset-expiration (:password-reset-expiration global-config 24)]

            (assoc context
              :db {:datasource db-conn-pool}
              :ds-realm ds-realm ; saved in case ds settings are udated
              :jdbc-realm jdbc-realm
              :host-private-key host-private-key
              :host-public-key host-public-key
              :sec-mgr sec-mgr
              :password-reset-expiration password-reset-expiration))))

  (stop [this context]
        (.close (get-in (service-context this) [:db :datasource]))
        context)

  (get-subject [this column value]
               (st/get-subject (:db (service-context this)) column value))

  (update-local-user-password! [this login new-password clear-token-uuid?]
                               (st/update-user-password! (:db (service-context this)) login (hash-password new-password) clear-token-uuid?))

  (password-matches? [this login current-password]
                     (-> (service-context this)
                         :jdbc-realm
                         (shiro/password-matches? login current-password)))

  (reset-password-token! [this requesting-login sid]
                         (let [reset-uuid (st/mark-user-for-reset (:db (service-context this)) sid)]
                           (create-password-reset-token (:host-private-key (service-context this))
                                                        (:password-reset-expiration (service-context this))
                                                        requesting-login
                                                        reset-uuid)))

  (validate-reset-token [this token]
                        (let [pub-key (:host-public-key (service-context this))]
                          (validate-token (service-context this) token pub-key)))

  (get-directory-settings [this subject]
                          (pm/get-directory-settings subject (:db (service-context this))))

  (update-directory-settings! [this subject data]
                              (let [svc-ctx (service-context this)]
                                (do-update-directory-settings! subject (:db svc-ctx) (:ds-realm svc-ctx) data)))

  (get-types [this]
             (st/get-types (:db (service-context this))))

  (session-id->subject [this session-id]
                       (shiro/session-id->subject (security-manager this) session-id))

  (login->subject [this login]
                  (shiro/login->subject (security-manager this) login))

  (security-manager [this]
                    (:sec-mgr (service-context this))))
